#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <math.h>
#include <chrono>

// 2 bucket radix search

void rng(uint64_t *arr, int n)
{
  int seed = 13516032;
  srand(seed);
  for (int i = 0; i < n; i++)
  {
    arr[i] = (uint64_t)rand();
  }
}

void print_arr(uint32_t* arr, int n)
{
	for (int i = 0; i < n; i++) {
		printf("%d ", arr[i]);
	}

	printf("\n");
}

__global__
void flag_array(uint64_t *arr, int n, int *flag, int idx) {
  int index = blockIdx.x * blockDim.x + threadIdx.x;
  int stride = blockDim.x * gridDim.x;

  for(uint64_t i = index; i < n; i += stride) {
    // flag[i] = (arr[i] & (1 << idx));
    flag[i] = (arr[i] >> idx) & 1;
  }
}

__global__
void count_hist(int *flag, int n, int* count) {
  int index = blockIdx.x * blockDim.x + threadIdx.x;
  int stride = blockDim.x * gridDim.x;

  for(int i = index; i < n; i += stride) {
    if(flag[i]) {
      atomicAdd(count + 1, 1);
    } else {
      atomicAdd(count, 1);
    }
  }
}

__global__
void inc_prefix_sum(int *inc_prefix, int n, int step, int *result_inc_prefix) {
  int dist = 1 << step;

  int index = blockIdx.x * blockDim.x + threadIdx.x;
  int stride = blockDim.x * gridDim.x;

  for(int i = index; i < n; i += stride) {
    result_inc_prefix[i] = inc_prefix[i];
    if(i - dist >= 0) {
      result_inc_prefix[i] += inc_prefix[i - dist];
    }
  }
}

__global__
void reverse_bit(int *dist, int *source, int size) {
  int index = blockIdx.x * blockDim.x + threadIdx.x;
  int stride = blockDim.x * gridDim.x;

  for(int i = index; i < size; i += stride)
    dist[i] = source[i] == 1 ? 0 : 1;
}

__global__
void calc_res_index(uint64_t *arr, int *flag, int *inc_prefix, int *rev_inc_prefix, int size, int *count, uint64_t *res_arr) {
  int index = blockIdx.x * blockDim.x + threadIdx.x;
  int stride = blockDim.x * gridDim.x;

  for(int i = index; i < size; i += stride) {
    int pref = 0;
    if(i > 0)
      if(flag[i] == 1) {
        pref = inc_prefix[i - 1];
      } else {
        pref = rev_inc_prefix[i - 1];
      }
    
    int c = flag[i] == 0 ? 0 : count[0];

    int res_index = c + pref;
    res_arr[res_index] = arr[i];
  }
}

__global__
void simple_copy(uint64_t *dist, uint64_t *source, int size) {
  int index = blockIdx.x * blockDim.x + threadIdx.x;
  int stride = blockDim.x * gridDim.x;

  for(int i = index; i < size; i += stride)
    dist[i] = source[i];
}

void radix_sort(uint64_t *arr, int n) {
  int block_size = 256;
  int num_block = (n + block_size - 1) / block_size;
  // int block_size = 1;
  // int num_block = 1;

  // determine largest item
  uint64_t max = arr[0];
  for(uint64_t i = 0; i < n; i++)
    if(arr[i] > max)
      max = arr[i];
    
  uint64_t* d_arr;
  cudaMallocManaged(&d_arr, n * sizeof(uint64_t));
  cudaMemcpy(d_arr, arr, n * sizeof(uint64_t), cudaMemcpyHostToHost);

  int* flag;
  cudaMallocManaged(&flag, n * sizeof(int));

  int* count;
  cudaMallocManaged(&count, 2 * sizeof(int));

  int* inc_prefix, *result_inc_prefix;
  cudaMallocManaged(&inc_prefix, n * sizeof(int));
  cudaMallocManaged(&result_inc_prefix, n * sizeof(int));

  int* rev_inc_prefix, *result_rev_inc_prefix;
  cudaMallocManaged(&rev_inc_prefix, n * sizeof(int));
  cudaMallocManaged(&result_rev_inc_prefix, n * sizeof(int));

  uint64_t* res_arr;
  cudaMallocManaged(&res_arr, n *sizeof(uint64_t));

  for(uint64_t idx = 0; ((1 << idx) <= max) && (idx < 64); idx += 1) {
    // computer flag and reverse flag
    flag_array<<<num_block, block_size>>>(d_arr, n, flag, idx);
    cudaDeviceSynchronize();

    cudaMemcpy(inc_prefix, flag, n * sizeof(int), cudaMemcpyDeviceToDevice);
    reverse_bit<<<num_block, block_size>>>(rev_inc_prefix, flag, n);
    cudaDeviceSynchronize();

    // count histogram from flag
    cudaMemset(count, 0, 2 * sizeof(int));
    count_hist<<<num_block, block_size>>>(flag, n, count);
    cudaDeviceSynchronize();


    // inclusive prefix sum
    // inc_prefix -> 1
    // rev_inc_prefix -> 0

    for(int step = 0; (1 << step) < n; step++) {
      inc_prefix_sum<<<num_block, block_size>>>(inc_prefix, n, step, result_inc_prefix);
      cudaDeviceSynchronize();
      
      inc_prefix_sum<<<num_block, block_size>>>(rev_inc_prefix, n, step, result_rev_inc_prefix);
      cudaDeviceSynchronize();

      cudaMemcpy(inc_prefix, result_inc_prefix, n * sizeof(int), cudaMemcpyDeviceToDevice);
      cudaMemcpy(rev_inc_prefix, result_rev_inc_prefix, n * sizeof(int), cudaMemcpyDeviceToDevice);
    }

    // for(int i = 0; i < n; i++)
    //   printf("i %d f %d %d\n", i, inc_prefix[i], rev_inc_prefix[i]);

    calc_res_index<<<num_block, block_size>>>(d_arr, flag, inc_prefix, rev_inc_prefix, n, count, res_arr);
    cudaDeviceSynchronize();
    // for(int i = 0; i < n; i++) {
    //   printf("i %d darr %d res %d\n", i, d_arr[i], res_arr[i]);
    // }
    simple_copy<<<num_block, block_size>>>(d_arr, res_arr, n);
    cudaDeviceSynchronize();
    // printf("===\n");
  }

  cudaMemcpy(arr, d_arr, n * sizeof(uint64_t), cudaMemcpyHostToHost);
  cudaFree(inc_prefix);
  cudaFree(res_arr);
  cudaFree(result_inc_prefix);
  cudaFree(rev_inc_prefix);
  cudaFree(result_rev_inc_prefix);
  cudaFree(d_arr);
  cudaFree(flag);
  cudaFree(count);
}

int main(int argc, char **argv)
{
  int N = atoi(argv[1]);
  uint64_t *arr;

  arr = (uint64_t*)malloc(N * sizeof(uint64_t));

  rng(arr, N);

  auto start_time = std::chrono::high_resolution_clock::now();
  radix_sort(arr, N);
  auto elapsed = std::chrono::high_resolution_clock::now() -start_time;
  auto microsecs = std::chrono::duration_cast<std::chrono::microseconds>(elapsed).count();

  for(int i = 0; i < N; i++) {
    printf("%d\n", arr[i]);
  }

  fprintf(stderr, "time: %lld us\n", microsecs);
  
  print_arr(arr, N);

  free(arr);
  return 0;
}